"""
Module with functions and script to generate tests directly
from our local version of gsw_check_functions.m.

Usage (run from this test directory):
    python check_functions.py

A primary use for this script is to see which functions are
missing from GSW-Python; they appear in the NameError category.

TypeError category can include functions that are incomplete
or otherwise not working correctly.

For functions that run but yield results that fail the check,
the error arrays are printed.

This can be improved--we should get less information from the
matlab script and more from our own functions.  We probably
should not need the matlab script at all, or maybe use it only
to extract the list of functions being tested in matlab.

This module is also imported by test_check_functions.py, which
is run by py.test.

"""

import os
import re
import sys

import numpy as np

from gsw import *  # noqa
from gsw._utilities import Bunch

# If we switch to using the logging module, uncomment:
# import logging
# log = logging.getLogger()
# logging.basicConfig()

#  The following re patterns are for the "alternative" parsing of
#  the test line to support using numpy assert_allclose.  This is
#  not presently in use, but aspects of this method, here and in
#  _WIP_test_ufuncs.py, might replace some of the original code here.
#
# pattern for a single test line after it has been pre-processed to
# remove spaces, the square brackets, and semicolon
testlinepat = r"(\w+\.\w+)=find\(\w*\((\w+\.\w+)-(\w+\.\w+)\)>=(\w+\.\w+)\)"
#
# pattern for the inner test when there is a sequence separated by '|'
testpat = r"\(+\w*\((\w+\.\w+)-(\w+\.\w+)\)+>=(\w+\.\w+)\)"


def find(x):
    """
    Numpy equivalent to Matlab find.
    """
    return np.nonzero(np.asarray(x).flatten())[0]


def group_or(line):
    """
    Translate matlab 'find' functions including logical or operators.

    Example: the test for gsw_rho_alpha_beta

    Numpy wart: using bitwise or as a fake elementwise logical or,
    we need to add parentheses.
    """
    if not ('find(' in line and '|' in line):
        return line
    i0 = line.index('find(') + 5
    head = line[:i0]
    tail = line[i0:]
    parts = tail.replace('|', ') | (')
    new = head + '(' + parts + ')'
    return new


class FunctionCheck:
    """
    Parse the line-pair for checks in gsw_check_functions.
    """
    def __init__(self, linepair):
        """
        *linepair* is the sequence of two lines; the first runs
        the function and assigns the output, and the second
        generates an array of indices where the output error
        exceeds a tolerance.
        """

        self.linepair = linepair
        self.runline = linepair[0]
        self.testline = linepair[1]

        # parse the line that runs the function
        head, tail = self.runline.split('=')
        self.outstrings = [s.strip() for s in head.split(',')]
        self.outstr = ','.join(self.outstrings)
        funcstr, argpart = tail.split('(', 1)
        self.name = funcstr.strip()
        self.argstrings = [s.strip() for s in argpart[:-1].split(',')]
        self.argstr = ','.join(self.argstrings)

        # parse the line that checks the results
        head, tail = self.testline.split('=', 1)
        self.resultstr = head.strip()  # cv.I*
        head, tail = tail.split('(', 1)
        self.teststr = tail.strip()[:-1]   # argument of "find()"
        self.teststr = self.teststr.replace('abs(', 'np.abs(')

        # alternative parsing of testline
        testline = self.testline.replace(' ', '')
        if '|' in testline:
            diffstr, test_str = testline.split('=', 1)
            # Chop off the leading 'find('.
            tests = test_str[5:].split('|')
            self.test_varstrings = [diffstr]
            for test in tests:
                m = re.match(testpat, test)
                if m is None:
                    print(self.name, testpat, test, m)
                if m is not None:
                    self.test_varstrings.extend(list(m.groups()))
        else:
            m = re.match(testlinepat, testline)
            if m is not None:
                self.test_varstrings =  m.groups()
            else:
                print("no match")
                self.test_varstrings = None


        # To be set when run() is successful
        self.outlist = None
        self.result = None   # will be a reference to the cv.I* array
        self.passed = None   # will be set to True or False

        # To be set if run() is not successful
        self.exception = None

    def __str__(self):
        return self.runline

    def record_details(self, evalargs):
        tline = self.testline
        i0 = 5 + tline.index('find(')
        tline = tline[i0:-1]
        checks = tline.split('|')
        parts = []
        for check in checks:
            check = check.replace(' ', '')
            if check.startswith('('):
                check = check[1:-1]
            part = Bunch(check=check)
            LHS, RHS = check.split('>=')
            part.tolerance = eval(RHS, *evalargs)
            # Sometimes there is an extra set of ().
            if LHS.startswith('('):
                LHS = LHS[1:-1]
            LHS = LHS[4:-1]  # chop off abs(...)
            target, calculated = LHS.split('-')
            part.checkval = eval(target, *evalargs)
            part.val = eval(calculated, *evalargs)
            parts.append(part)

        self.details = parts

    def run(self, locals_=None):
        try:
            if locals_ is not None:
                _globals = globals() #dict(**globals())
                _globals.update(locals_)
                evalargs = (_globals,)
            else:
                evalargs = ()

            # The following is needed for melting_ice_into_seawater.
            if len(self.outstrings) > 1:
                rl_ind = f"[:{len(self.outstrings)}]"
            else:
                rl_ind = ""

            exec(self.runline + rl_ind, *evalargs)
            if len(self.outstrings) == 1:
                if isinstance(eval(self.outstr, *evalargs), tuple):
                    exec(f"{self.outstr} = {self.outstr}[0]", *evalargs)
            self.outlist = [eval(s, *evalargs) for s in self.outstrings]

            exec(self.testline, *evalargs)
            self.result = eval(self.resultstr, *evalargs)

            self.passed = (len(self.result) == 0)
            # The following has trouble with CT_first_derivatives
            if self.name not in ['CT_first_derivatives',]:
                self.record_details(evalargs)
            # print("%s passed? %s" % (self.name, self.passed))

        except Exception as e:
            self.exception = e


def find_arguments(checks):
    """
    For a sequence of FunctionCheck instances, return the
    set of unique arguments as a sorted list.
    """
    argset = set()
    for c in checks:
        argset.update(c.argstrings)
    argsetlist = list(argset)
    argsetlist.sort()
    return argsetlist


def find_arglists(checks):
    """
    For a sequence of FunctionCheck instances, return the
    set of unique argument lists as a sorted list.
    """
    alset = set()
    for c in checks:
        alset.update([c.argstr])
    arglists = list(alset)
    arglists.sort()
    return arglists

def parse_check_functions(mfile):
    """
    Return a list of FunctionCheck instances from gsw_check_functions.m
    """

    with open(mfile) as fid:
        mfilelines = fid.readlines()

    first_pass = []

    concat = False
    for line in mfilelines:
        line = line.strip()
        if concat:
            if line.endswith('...'):
                line = line[:-3]
            first_pass[-1] += line
            if line.endswith(';'):
                concat = False
            continue
        if '=' in line and (line.startswith('gsw_') or line.startswith('[gsw_')):
            if line.endswith('...'):
                line = line[:-3]
                concat = True
            first_pass.append(line)

    second_pass = []

    for line in first_pass:
        if not '(' in line:
            continue
        if 'which' in line:
            continue
        line = line.replace('gsw_', '')
        if line.startswith('['):
            line = line[1:].replace(']', '')
        if line.endswith(';'):
            line = line[:-1]
        line = line.replace('(I)', '')  # For deltaSA_atlas.
        second_pass.append(line)

    pairs = []

    for i in range(len(second_pass)):
        if 'find(' in second_pass[i] and not 'find(' in second_pass[i-1]:
            pairs.extend(second_pass[i-1:i+1])

    final = [group_or(line) for line in pairs]

    checks = []
    for i in range(0, len(final), 2):
        pair = final[i:i+2]
        checks.append(FunctionCheck(pair))

    return checks

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
                description='Run checks from gsw_check_functions.m')

    parser.add_argument('--path', dest='mfiledir',
                        default="",
                       help='path to external gsw_check_functions.m')
    parser.add_argument('-v', '--verbose',
                        action='store_true',
                        help='print output mismatch arrays')
    parser.add_argument('--find',
                        help='run functions with this substring')

    args = parser.parse_args()

    if args.mfiledir:
        mfile = os.path.join(args.mfiledir, "gsw_check_functions.m")
    else:
        mfile = "gsw_check_functions_save.m"
    checks = parse_check_functions(mfile)

    #datadir = os.path.join(os.path.dirname(gsw.utilities.__file__), 'data')
    datadir = './'
    cv = Bunch(np.load(os.path.join(datadir, 'gsw_cv_v3_0.npz')))
    cf = Bunch()

    if args.find:
        checks = [c for c in checks if args.find in c.runline]

    for fc in checks:
        fc.run()

    passes = [f for f in checks if f.passed]
    failures = [f for f in checks if f.passed is False]

    run_problems = [f for f in checks if f.exception is not None]

    etypes = [NameError, UnboundLocalError, TypeError, AttributeError]
    ex_dict = {}
    for exc in etypes:
        elist = [(f.name, f.exception) for f in checks if
                 isinstance(f.exception, exc)]
        ex_dict[exc] = elist

    print(f"\n{len(checks)} tests were translated from gsw_check_functions.m")
    print(f"\n{len(passes)} tests ran with no error and with correct output")
    if args.verbose:
        for f in passes:
            print(f.name)

    print(f"\n{len(failures)} tests had an output mismatch:")
    for f in failures:
        print(f.name)
        print(f.runline)
        print(f.testline)
        if args.verbose:
            print(f.result)
            for part in f.details:
                print("tolerance: ", part.tolerance)
                print("error:")
                print(part.checkval - part.val)
                print('')

        print('')

    print(f"\n{len(run_problems)} exceptions were raised as follows:")
    for exc in etypes:
        print("  ", exc.__name__)
        strings = ["     {} : {}".format(*e) for e in ex_dict[exc]]
        print("\n".join(strings))
        print("")

    checkbunch = Bunch([(c.name, c) for c in checks])



"""
Functions for vertical interpolation.
"""

import numpy as np

from . import _gsw_ufuncs
from ._utilities import indexer, match_args_return

__all__ = ['sa_ct_interp',
           'tracer_ct_interp',
           ]

@match_args_return
def sa_ct_interp(SA, CT, p, p_i, axis=0):
    """
    Interpolates vertical casts of values of Absolute Salinity
    and Conservative Temperature to the arbitrary pressures p_i.

    Parameters
    ----------
    SA : array-like
        Absolute Salinity, g/kg
    CT : array-like
        Conservative Temperature (ITS-90), degrees C
    p : array-like
        Sea pressure (absolute pressure minus 10.1325 dbar), dbar
    p_i : array-like
        Sea pressure to interpolate on, dbar
    axis : int, optional, default is 0
        The index of the pressure dimension in SA and CT.


    Returns
    -------
    SA_i : array
        Values of SA interpolated to p_i along the specified axis.
    CT_i : array
        Values of CT interpolated to p_i along the specified axis.

    """
    if SA.shape != CT.shape:
        raise ValueError(f'Shapes of SA and CT must match; found {SA.shape} and {CT.shape}')
    if p.ndim != p_i.ndim:
        raise ValueError(f'p and p_i must have the same number of dimensions;\n'
                         f' found {p.ndim} versus {p_i.ndim}')
    if p.ndim == 1 and SA.ndim > 1:
        if len(p) != SA.shape[axis]:
            raise ValueError(
                f'With 1-D p, len(p) must be SA.shape[axis];\n'
                f' found {len(p)} versus {SA.shape[axis]} on specified axis, {axis}'
                )
        ind = [np.newaxis] * SA.ndim
        ind[axis] = slice(None)
        p = p[tuple(ind)]
        p_i = p_i[tuple(ind)]
    elif p.ndim > 1:
        if p.shape != SA.shape:
            raise ValueError(f'With {p.ndim}-D p, shapes of p and SA must match;\n'
                             f'found {p.shape} and {SA.shape}')
        if any(p.shape[i] != p_i.shape[i] for i in range(p.ndim) if i != axis):
            raise ValueError(f'With {p.ndim}-D p, p and p_i must have the same dimensions outside of axis {axis};\n'
                             f' found {p.shape} versus {p_i.shape}')
    with np.errstate(invalid='ignore'):
        # The need for this context seems to be a bug in np.ma.any.
        if np.ma.any(np.ma.diff(np.ma.masked_invalid(p_i), axis=axis) <= 0) \
                or np.ma.any(np.ma.diff(np.ma.masked_invalid(p), axis=axis) <= 0):
            raise ValueError('p and p_i must be increasing along the specified axis')
    p = np.broadcast_to(p, SA.shape)
    goodmask = ~(np.isnan(SA) | np.isnan(CT) | np.isnan(p))
    SA_i = np.empty(p_i.shape, dtype=float)
    CT_i = np.empty(p_i.shape, dtype=float)
    SA_i.fill(np.nan)
    CT_i.fill(np.nan)

    try:
        order = 'F' if SA.flags.fortran else 'C'
    except AttributeError:
        order = 'C'  # e.g., xarray DataArray doesn't have flags
    for ind in indexer(SA.shape, axis, order=order):
        # this is needed to support xarray inputs for numpy < 1.23
        igood = np.asarray(goodmask[ind])
        pgood = p[ind][igood]
        pi = p_i[ind]
        # There must be at least 2 non-NaN values for interpolation
        if len(pgood) > 2:
            sa = SA[ind][igood]
            ct = CT[ind][igood]
            sai, cti = _gsw_ufuncs.sa_ct_interp(sa, ct, pgood, pi)
            SA_i[ind] = sai
            CT_i[ind] = cti

    return (SA_i, CT_i)

@match_args_return
def tracer_ct_interp(tracer, CT, p, p_i, factor=9., axis=0):
    """
    Interpolates vertical casts of values of a tracer
    and Conservative Temperature to the arbitrary pressures p_i.

    Parameters
    ----------
    tracer : array-like
        tracer
    CT : array-like
        Conservative Temperature (ITS-90), degrees C
    p : array-like
        Sea pressure (absolute pressure minus 10.1325 dbar), dbar
    p_i : array-like
        Sea pressure to interpolate on, dbar
    factor: float, optional, default is 9.
        Ratio between the ranges of Conservative Temperature
        and tracer in the world ocean.
    axis : int, optional, default is 0
        The index of the pressure dimension in tracer and CT.


    Returns
    -------
    tracer_i : array
        Values of tracer interpolated to p_i along the specified axis.
    CT_i : array
        Values of CT interpolated to p_i along the specified axis.

    """
    if tracer.shape != CT.shape:
        raise ValueError(f'Shapes of tracer and CT must match; found {tracer.shape} and {CT.shape}')
    if p.ndim != p_i.ndim:
        raise ValueError(f'p and p_i must have the same number of dimensions;\n'
                         f' found {p.ndim} versus {p_i.ndim}')
    if p.ndim == 1 and tracer.ndim > 1:
        if len(p) != tracer.shape[axis]:
            raise ValueError(
                f'With 1-D p, len(p) must be tracer.shape[axis];\n'
                f' found {len(p)} versus {tracer.shape[axis]} on specified axis, {axis}'
                )
        ind = [np.newaxis] * tracer.ndim
        ind[axis] = slice(None)
        p = p[tuple(ind)]
        p_i = p_i[tuple(ind)]
    elif p.ndim > 1:
        if p.shape != tracer.shape:
            raise ValueError(f'With {p.ndim}-D p, shapes of p and tracer must match;\n'
                             f'found {p.shape} and {tracer.shape}')
        if any(p.shape[i] != p_i.shape[i] for i in range(p.ndim) if i != axis):
            raise ValueError(f'With {p.ndim}-D p, p and p_i must have the same dimensions outside of axis {axis};\n'
                             f' found {p.shape} versus {p_i.shape}')
    with np.errstate(invalid='ignore'):
        # The need for this context seems to be a bug in np.ma.any.
        if np.ma.any(np.ma.diff(np.ma.masked_invalid(p_i), axis=axis) <= 0) \
                or np.ma.any(np.ma.diff(np.ma.masked_invalid(p), axis=axis) <= 0):
            raise ValueError('p and p_i must be increasing along the specified axis')
    p = np.broadcast_to(p, tracer.shape)
    goodmask = ~(np.isnan(tracer) | np.isnan(CT) | np.isnan(p))
    tracer_i = np.empty(p_i.shape, dtype=float)
    CT_i = np.empty(p_i.shape, dtype=float)
    tracer_i.fill(np.nan)
    CT_i.fill(np.nan)

    try:
        order = 'F' if tracer.flags.fortran else 'C'
    except AttributeError:
        order = 'C'  # e.g., xarray DataArray doesn't have flags
    for ind in indexer(tracer.shape, axis, order=order):
        # this is needed to support xarray inputs for numpy < 1.23
        igood = np.asarray(goodmask[ind])
        pgood = p[ind][igood]
        pi = p_i[ind]
        # There must be at least 2 non-NaN values for interpolation
        if len(pgood) > 2:
            tr = tracer[ind][igood]
            ct = CT[ind][igood]
            tri, cti = _gsw_ufuncs.tracer_ct_interp(tr, ct, pgood, pi, factor)
            tracer_i[ind] = tri
            CT_i[ind] = cti

    return (tracer_i, CT_i)



"""
Tests functions with xarray inputs.

This version is a copy of the original test_check_functions but with
an import of xarray, and conversion of the 3 main check cast arrays
into DataArray objects.

An additional xarray-dask test is added.
"""

import os

import numpy as np
import pandas as pd
import pytest
from check_functions import parse_check_functions
from numpy.testing import assert_allclose

import gsw
from gsw._utilities import Bunch

xr = pytest.importorskip('xarray')

# Most of the tests have some nan values, so we need to suppress the warning.
# Any more careful fix would likely require considerable effort.
np.seterr(invalid='ignore')

root_path = os.path.abspath(os.path.dirname(__file__))

# Function checks that we can't handle automatically yet.
blacklist = ['deltaSA_atlas',  # the test is complicated; doesn't fit the pattern.
             'geostrophic_velocity',  # test elsewhere; we changed the API
             #'CT_from_entropy', # needs prior entropy_from_CT; don't have it in C
             #'CT_first_derivatives', # passes, but has trouble in "details";
                                      # see check_functions.py
             #'entropy_second_derivatives', # OK now; handling extra parens.
             #'melting_ice_into_seawater',  # OK now; fixed nargs mismatch.
             ]

# We get an overflow from ct_from_enthalpy_exact, but the test passes.
cv = Bunch(np.load(os.path.join(root_path, 'gsw_cv_v3_0.npz')))

# Substitute new check values for the pchip interpolation version.
cv.geo_strf_dyn_height = np.load(os.path.join(root_path,'geo_strf_dyn_height.npy'))
cv.geo_strf_velocity = np.load(os.path.join(root_path,'geo_strf_velocity.npy'))

for name in ['SA_chck_cast', 't_chck_cast', 'p_chck_cast']:
    cv[name] = xr.DataArray(cv[name])

cf = Bunch()

d = dir(gsw)
funcnames = [name for name in d if '__' not in name]

mfuncs = parse_check_functions(os.path.join(root_path, 'gsw_check_functions_save.m'))
mfuncs = [mf for mf in mfuncs if mf.name in d and mf.name not in blacklist]
mfuncnames = [mf.name for mf in mfuncs]


@pytest.fixture(scope='session', params=mfuncs)
def cfcf(request):
    return cv, cf, request.param


def test_check_function(cfcf):
    cv, cf, mfunc = cfcf
    mfunc.run(locals())
    if mfunc.exception is not None or not mfunc.passed:
        print('\n', mfunc.name)
        print('  ', mfunc.runline)
        print('  ', mfunc.testline)
        if mfunc.exception is None:
            mfunc.exception = ValueError('Calculated values are different from the expected matlab results.')
        raise mfunc.exception
    else:
        print(mfunc.name)
        assert mfunc.passed


def test_dask_chunking():
    dsa = pytest.importorskip('dask.array')

    # define some input data
    shape = (100, 1000)
    chunks = (100, 200)
    sp = xr.DataArray(dsa.full(shape, 35., chunks=chunks), dims=['time', 'depth'])
    p = xr.DataArray(np.arange(shape[1]), dims=['depth'])
    lon = 0
    lat = 45

    sa = gsw.SA_from_SP(sp, p, lon, lat)
    sa_dask = sa.compute()

    sa_numpy = gsw.SA_from_SP(np.full(shape, 35.0), p.values, lon, lat)
    assert_allclose(sa_dask, sa_numpy)


# Additional tests from Graeme MacGilchrist
# https://nbviewer.jupyter.org/github/gmacgilchrist/wmt_bgc/blob/master/notebooks/test_gsw-xarray.ipynb

# Define dimensions and coordinates
dims = ['y','z','t']
# 2x2x2
y = np.arange(0,2)
z = np.arange(0,2)
t = np.arange(0,2)
# Define numpy arrays of salinity, temperature and pressure
SA_vals = np.array([[[34.7,34.8],[34.9,35]],[[35.1,35.2],[35.3,35.4]]])
CT_vals = np.array([[[7,8],[9,10]],[[11,12],[13,14]]])
p_vals = np.array([10,20])
lat_vals = np.array([0,10])
# Plug in to xarray objects
SA = xr.DataArray(SA_vals,dims=dims,coords={'y':y,'z':z,'t':t})
CT = xr.DataArray(CT_vals,dims=dims,coords={'y':y,'z':z,'t':t})
p = xr.DataArray(p_vals,dims=['z'],coords={'z':z})
lat = xr.DataArray(lat_vals,dims=['y'],coords={'y':y})


def test_xarray_with_coords():
    pytest.importorskip('dask')
    SA_chunk = SA.chunk(chunks={'y':1,'t':1})
    CT_chunk = CT.chunk(chunks={'y':1,'t':1})
    lat_chunk = lat.chunk(chunks={'y':1})

    # Dimensions and coordinates match:
    expected = gsw.sigma0(SA_vals, CT_vals)
    xarray = gsw.sigma0(SA, CT)
    chunked = gsw.sigma0(SA_chunk, CT_chunk)
    assert_allclose(xarray, expected)
    assert_allclose(chunked, expected)

    # Broadcasting along dimension required (dimensions known)
    expected = gsw.alpha(SA_vals, CT_vals, p_vals[np.newaxis, :, np.newaxis])
    xarray = gsw.alpha(SA, CT, p)
    chunked = gsw.alpha(SA_chunk, CT_chunk, p)
    assert_allclose(xarray, expected)
    assert_allclose(chunked, expected)

    # Broadcasting along dimension required (dimensions unknown/exclusive)
    expected = gsw.z_from_p(p_vals[:, np.newaxis], lat_vals[np.newaxis, :])
    xarray = gsw.z_from_p(p, lat)
    chunked = gsw.z_from_p(p,lat_chunk)
    assert_allclose(xarray, expected)
    assert_allclose(chunked, expected)

def test_pandas_20():
    df = pd.DataFrame(
        {
            "pressure": [0, 10, 20],
            "latitude": [70, 70, 70],
        }
    )
    depth = -1 * gsw.z_from_p(df["pressure"], df["latitude"])
    assert isinstance(depth, pd.core.series.Series)


import ctypes

import gsw


def test_ctypes_access():
    dllname = gsw._gsw_ufuncs.__file__
    gswlib = ctypes.cdll.LoadLibrary(dllname)
    rho_gsw_ctypes = gswlib.gsw_rho  # In-situ density.
    rho_gsw_ctypes.argtypes = [ctypes.c_double] * 3
    rho_gsw_ctypes.restype = ctypes.c_double
    stp = (35.0, 10.0, 0.0)
    assert rho_gsw_ctypes(*stp) == gsw.rho(*stp)



"""
Minimal setup.py for building gswc.
"""


import os
import shutil
import sys

import numpy
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext as _build_ext

rootpath = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    return open(os.path.join(rootpath, *parts)).read()


class build_ext(_build_ext):
    # Extension builder from pandas without the cython stuff
    def build_extensions(self):
        numpy_incl = numpy.get_include()

        for ext in self.extensions:
            if hasattr(ext, "include_dirs") and not numpy_incl in ext.include_dirs:
                ext.include_dirs.append(numpy_incl)
        _build_ext.build_extensions(self)


# MSVC can't handle C complex, and distutils doesn't seem to be able to
# let us force C++ compilation of .c files, so we use the following hack for
# Windows.
if sys.platform == "win32":
    cext = "cpp"
    shutil.copy(
        "src/c_gsw/gsw_oceanographic_toolbox.c",
        "src/c_gsw/gsw_oceanographic_toolbox.cpp",
    )
    shutil.copy("src/c_gsw/gsw_saar.c", "src/c_gsw/gsw_saar.cpp")
else:
    cext = "c"

ufunc_src_list = [
    "src/_ufuncs.c",
    "src/c_gsw/gsw_oceanographic_toolbox." + cext,
    "src/c_gsw/gsw_saar." + cext,
]

config = {
    "ext_modules": [Extension("gsw._gsw_ufuncs", ufunc_src_list)],
    "include_dirs": [os.path.join(rootpath, "src", "c_gsw")],
    "cmdclass": {"build_ext": build_ext},
}

setup(**config)



import os

import numpy as np
from numpy.testing import assert_allclose

import gsw
from gsw._utilities import Bunch

root_path = os.path.abspath(os.path.dirname(__file__))

cv = Bunch(np.load(os.path.join(root_path, 'gsw_cv_v3_0.npz')))

def test_sa_ct_interp():
    p = cv.p_chck_cast
    CT = cv.CT_chck_cast
    SA = cv.SA_chck_cast
    p_i = np.repeat(cv.p_i[:, np.newaxis], p.shape[1], axis=1)
    SA_i, CT_i = gsw.sa_ct_interp(SA, CT, p, p_i)
    assert_allclose(SA_i, cv.SAi_SACTinterp, rtol=0, atol=cv.SAi_SACTinterp_ca)
    assert_allclose(CT_i, cv.CTi_SACTinterp, rtol=0, atol=cv.CTi_SACTinterp_ca)

def test_tracer_ct_interp():
    p = cv.p_chck_cast
    CT = cv.CT_chck_cast
    tracer = cv.SA_chck_cast
    p_i = np.repeat(cv.p_i[:, np.newaxis], p.shape[1], axis=1)
    tracer_i, CT_i = gsw.tracer_ct_interp(tracer, CT, p, p_i)
    assert_allclose(tracer_i, cv.traceri_tracerCTinterp, rtol=0, atol=cv.traceri_tracerCTinterp_ca)
    assert_allclose(CT_i, cv.CTi_SACTinterp, rtol=0, atol=cv.CTi_SACTinterp_ca)



"""
Functions for calculating geostrophic currents.
"""

import numpy as np

from . import _gsw_ufuncs
from ._utilities import indexer, match_args_return
from .conversions import z_from_p

__all__ = ['geo_strf_dyn_height',
           'distance',
           'f',
           'geostrophic_velocity',
           ]

@match_args_return
def geo_strf_dyn_height(SA, CT, p, p_ref=0, axis=0, max_dp=1.0,
                        interp_method='pchip'):
    """
    Dynamic height anomaly as a function of pressure.

    Parameters
    ----------
    SA : array-like
        Absolute Salinity, g/kg
    CT : array-like
        Conservative Temperature (ITS-90), degrees C
    p : array-like
        Sea pressure (absolute pressure minus 10.1325 dbar), dbar
    p_ref : float or array-like, optional
        Reference pressure, dbar
    axis : int, optional, default is 0
        The index of the pressure dimension in SA and CT.
    max_dp : float
        If any pressure interval in the input p exceeds max_dp, the dynamic
        height will be calculated after interpolating to a grid with this
        spacing.
    interp_method : string {'mrst', 'pchip', 'linear'}
        Interpolation algorithm.

    Returns
    -------
    dynamic_height : array
        This is the integral of specific volume anomaly with respect
        to pressure, from each pressure in p to the specified
        reference pressure.  It is the geostrophic streamfunction
        in an isobaric surface, relative to the reference surface.

    """
    interp_methods = {'mrst' : 3, 'pchip' : 2, 'linear' : 1}
    if interp_method not in interp_methods:
        raise ValueError(f'interp_method must be one of {interp_methods.keys()}')
    if SA.shape != CT.shape:
        raise ValueError(f'Shapes of SA and CT must match; found {SA.shape} and {CT.shape}')
    if p.ndim == 1 and SA.ndim > 1:
        if len(p) != SA.shape[axis]:
            raise ValueError(
                f'With 1-D p, len(p) must be SA.shape[axis];\n'
                f' found {len(p)} versus {SA.shape[axis]} on specified axis, {axis}'
                )
        ind = [np.newaxis] * SA.ndim
        ind[axis] = slice(None)
        p = p[tuple(ind)]
    p_ref = float(p_ref)
    with np.errstate(invalid='ignore'):
        # The need for this context seems to be a bug in np.ma.any.
        if np.ma.any(np.ma.diff(np.ma.masked_invalid(p), axis=axis) <= 0):
            raise ValueError('p must be increasing along the specified axis')
    p = np.broadcast_to(p, SA.shape)
    goodmask = ~(np.isnan(SA) | np.isnan(CT) | np.isnan(p))
    dh = np.empty(SA.shape, dtype=float)
    dh.fill(np.nan)

    try:
        order = 'F' if SA.flags.fortran else 'C'
    except AttributeError:
        order = 'C'  # e.g., xarray DataArray doesn't have flags
    for ind in indexer(SA.shape, axis, order=order):
        # this is needed to support xarray inputs for numpy < 1.23
        igood = np.asarray(goodmask[ind])
        # If p_ref is below the deepest value, skip the profile.
        pgood = p[ind][igood]
        if  len(pgood) > 1 and pgood[-1] >= p_ref:
            sa = SA[ind][igood]
            ct = CT[ind][igood]
            # Temporarily add a top (typically surface) point and mixed layer
            # if p_ref is above the shallowest pressure.
            if pgood[0] > p_ref:
                ptop = np.arange(p_ref, pgood[0], max_dp)
                ntop = len(ptop)
                sa = np.hstack(([sa[0]]*ntop, sa))
                ct = np.hstack(([ct[0]]*ntop, ct))
                pgood = np.hstack((ptop, pgood))
            else:
                ntop = 0
            dh_all = _gsw_ufuncs.geo_strf_dyn_height_1(
                                         sa, ct, pgood, p_ref, max_dp,
                                         interp_methods[interp_method])
            if ntop > 0:
                dh[ind][igood] = dh_all[ntop:]
            else:
                dh[ind][igood] = dh_all

    return dh


def unwrap(lon, centered=True, copy=True):
    """
    Unwrap a sequence of longitudes or headings in degrees.

    Optionally center it as close to zero as possible

    By default, return a copy; if *copy* is False, avoid a
    copy when possible.

    Returns a masked array only if the input is a masked array.
    """
    # From pycurrents.data.ocean.  It could probably be simplified
    # for use here.

    masked_input = np.ma.isMaskedArray(lon)
    if masked_input:
        fill_value = lon.fill_value
        # masked_invalid loses the original fill_value (ma bug, 2011/01/20)
    lon = np.ma.masked_invalid(lon).astype(float)
    if lon.ndim != 1:
        raise ValueError("Only 1-D sequences are supported")
    if lon.shape[0] < 2:
        return lon
    x = lon.compressed()
    if len(x) < 2:
        return lon
    w = np.zeros(x.shape[0]-1, int)
    ld = np.diff(x)
    np.putmask(w, ld > 180, -1)
    np.putmask(w, ld < -180, 1)
    x[1:] += (w.cumsum() * 360.0)

    if centered:
        x -= 360 * np.round(x.mean() / 360.0)

    if lon.mask is np.ma.nomask:
        lon[:] = x
    else:
        lon[~lon.mask] = x
    if masked_input:
        lon.fill_value = fill_value
        return lon
    else:
        return lon.filled(np.nan)


@match_args_return
def distance(lon, lat, p=0, axis=-1):
    """
    Great-circle distance in m between lon, lat points.

    Parameters
    ----------
    lon, lat : array-like, 1-D or 2-D (shapes must match)
        Longitude, latitude, in degrees.
    p : array-like, scalar, 1-D or 2-D, optional, default is 0
        Sea pressure (absolute pressure minus 10.1325 dbar), dbar
    axis : int, -1, 0, 1, optional
        The axis or dimension along which *lat and lon* vary.
        This differs from most functions, for which axis is the
        dimension along which p increases.

    Returns
    -------
    distance : 1-D or 2-D array
        distance in meters between adjacent points.

    """
    earth_radius = 6371e3

    if not lon.shape == lat.shape:
        raise ValueError(f'lon, lat shapes must match; found {lon.shape}, {lat.shape}')
    if not (lon.ndim in (1, 2) and lon.shape[axis] > 1):
        raise ValueError('lon, lat must be 1-D or 2-D with more than one point'
                         f' along axis; found shape {lon.shape} and axis {axis}')
    if lon.ndim == 1:
        one_d = True
        # xarray requires expand_dims() rather than [newaxis, :]
        lon = np.expand_dims(lon, 0)
        lat = np.expand_dims(lat, 0)
        axis = -1
    else:
        one_d = False

    # Handle scalar default; match_args_return doesn't see it.
    p = np.atleast_1d(p)
    one_d = (one_d and p.ndim == 1)

    if axis == 0:
        indm = (slice(0, -1), slice(None))
        indp = (slice(1, None), slice(None))
    else:
        indm = (slice(None), slice(0, -1))
        indp = (slice(None), slice(1, None))

    if np.all(p == 0):
        z = 0
    else:
        lon, lat, p = np.broadcast_arrays(lon, lat, p)

        p_mid = 0.5 * (p[indm] + p[indp])
        lat_mid = 0.5 * (lat[indm] + lat[indp])

        z = z_from_p(p_mid, lat_mid)

    lon = np.radians(lon)
    lat = np.radians(lat)

    dlon = np.diff(lon, axis=axis)
    dlat = np.diff(lat, axis=axis)

    a = ((np.sin(dlat / 2)) ** 2 + np.cos(lat[indm]) *
         np.cos(lat[indp]) * (np.sin(dlon / 2)) ** 2)

    angles = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

    distance = (earth_radius + z) * angles

    if one_d:
        distance = distance[0]

    return distance


@match_args_return
def f(lat):
    """
    Coriolis parameter in 1/s for latitude in degrees.
    """
    omega = 7.292115e-5  # (1/s)   (Groten, 2004).
    f = 2 * omega * np.sin(np.radians(lat))
    return f

@match_args_return
def geostrophic_velocity(geo_strf, lon, lat, p=0, axis=0):
    """
    Calculate geostrophic velocity from a streamfunction.

    Calculates geostrophic velocity relative to a reference pressure,
    given a geostrophic streamfunction and the position of each station
    in sequence along an ocean section.  The data can be from a single
    isobaric or "density" surface, or from a series of such surfaces.

    Parameters
    ----------
    geo_strf : array-like, 1-D or 2-D
        geostrophic streamfunction; see Notes below.
    lon : array-like, 1-D
        Longitude, -360 to 360 degrees
    lat : array-like, 1-D
        Latitude, degrees
    p : float or array-like, optional
        Sea pressure (absolute pressure minus 10.1325 dbar), dbar.
        This used only for a tiny correction in the distance calculation;
        it is safe to omit it.
    axis : int, 0 or 1, optional
        The axis or dimension along which pressure increases in geo_strf.
        If geo_strf is 1-D, it is ignored.

    Returns
    -------
    velocity : array, 2-D or 1-D
        Geostrophic velocity in m/s relative to the sea surface,
        averaged between each successive pair of positions.
    mid_lon, mid_lat : array, 1-D
        Midpoints of input lon and lat.

    Notes
    -----
    The geostrophic streamfunction can be:

    - geo_strf_dyn_height (in an isobaric surface)
    - geo_strf_Montgomery (in a specific volume anomaly surface)
    - geo_strf_Cunninhgam (in an approximately neutral surface
      such as a potential density surface).
    - geo_strf_isopycnal (in an approximately neutral surface
      such as a potential density surface, a Neutral Density
      surface, or an omega surface (Klocker et al., 2009)).

    Only :func:`geo_strf_dyn_height` is presently implemented
    in GSW-Python.

    """
    lon = unwrap(lon)

    if lon.shape != lat.shape or lon.ndim != 1:
        raise ValueError(f'lon, lat must be 1-D and matching; found shapes {lon.shape} and {lat.shape}')

    if geo_strf.ndim not in (1, 2):
        raise ValueError(f'geo_strf must be 1-D or 2-d; found shape {geo_strf.shape}')

    laxis = 0 if axis else -1

    ds = distance(lon, lat, p)

    mid_lon = 0.5 * (lon[:-1] + lon[1:])
    mid_lat = 0.5 * (lat[:-1] + lat[1:])

    u = np.diff(geo_strf, axis=laxis) / (ds * f(mid_lat))

    return u, mid_lon, mid_lat
